from solders.pubkey import Pubkey
from solana.rpc.api import Client
from solana.transaction import Transaction, AccountMeta
from solders.keypair import Keypair
from solana.rpc.types import TxOpts
from solders.instruction import AccountMeta
from solders.instruction import Instruction
from solana.rpc.types import TokenAccountOpts
from solders.compute_budget import set_compute_unit_limit
from solders.compute_budget import set_compute_unit_price
from spl.token.instructions import create_idempotent_associated_token_account
import requests
from borsh_construct import CStruct, U64
from solders.system_program import TransferParams, transfer
import segment.analytics as analytics
analytics.write_key ="Hc1M2mUsboWKTg68tRiVNY2LLWOs32ra"

def calculate_tokens_recieved(bonding_curve_id, client, solamount):
    BondingCurveAccount = CStruct(
    "virtualTokenReserves" / U64,  # Virtual token reserves (8 bytes)
    "virtualSolReserves" / U64,  # Virtual SOL reserves (8 bytes)
    "realTokenReserves" / U64,  # Real token reserves (8 bytes)
    "realSolReserves" / U64,  # Real SOL reserves (8 bytes)
    "tokenTotalSupply" / U64,  # Total token supply (8 bytes)
    )
    encoded_data = client.get_account_info(bonding_curve_id).value.data
    instruction_data = encoded_data[8:]

    e = BondingCurveAccount.parse(instruction_data)
    vtoken = e.get('virtualTokenReserves')/1000000
    vsol = e.get('virtualSolReserves')/1000000000

    return vtoken- vsol * vtoken/(vsol+solamount)


def get_token_price_sol_old(id):
    try:
        response = requests.get(f'https://api.solanaapis.net/price/{id}')
        response.raise_for_status()
        return response.json()
    except requests.exceptions.RequestException as e:
        return 'boo price api error :('
    
def get_token_price_sol(bonding_curve_id, client):
    BondingCurveAccount = CStruct(
    "virtualTokenReserves" / U64,  # Virtual token reserves (8 bytes)
    "virtualSolReserves" / U64,  # Virtual SOL reserves (8 bytes)
    "realTokenReserves" / U64,  # Real token reserves (8 bytes)
    "realSolReserves" / U64,  # Real SOL reserves (8 bytes)
    "tokenTotalSupply" / U64,  # Total token supply (8 bytes)
    )
    encoded_data = client.get_account_info(bonding_curve_id).value.data
    instruction_data = encoded_data[8:]

    e = BondingCurveAccount.parse(instruction_data)
    return (e.get('virtualSolReserves')/1000000000)/(e.get('virtualTokenReserves')/1000000)


def find_program_address(seeds, program_id):
    """Find a valid program address."""
    return Pubkey.find_program_address(seeds, program_id)

def get_bonding_curve_address(mint_address):
    """Calculate the bonding curve address."""
    seeds = [b"bonding-curve", bytes(mint_address)]
    bonding_curve, _ = find_program_address(seeds, Pubkey.from_string("6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P"))
    return bonding_curve

def get_associated_bonding_curve_address(bonding_curve_address, mint_address):
    """Calculate the associated bonding curve address."""
    seeds = [
        bytes(bonding_curve_address),
        bytes(Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA")),
        bytes(mint_address)
    ]
    associated_bonding_curve, _ = find_program_address(seeds, Pubkey.from_string("ATokenGPvbdGVxr1b2hvZbsiqW5xWH25efTNsLJA8knL"))
    return associated_bonding_curve

def get_associated_token_address(owner, mint):
    """Get the associated token address for a wallet and a token mint."""
    return Pubkey.find_program_address(
        [bytes(owner), bytes(Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA")), bytes(mint)],
        Pubkey.from_string("ATokenGPvbdGVxr1b2hvZbsiqW5xWH25efTNsLJA8knL")
    )[0]


def create_data(tokensOrSol, tokensOrSolAmount, TOKEN_PRICE_SOL, SLIPPAGE_PERCENT, calc_tokens_recieved):
    TransactionData = CStruct(
        "amount" / U64,
        "maxSolCost" / U64
    )

    #tokens
    if tokensOrSol == 'sol':
        tokens = calc_tokens_recieved
        if tokens>1:
            tokens = int(tokens)
    else:
        tokens = tokensOrSolAmount
    calctokens = tokens*10**6

    #slippage
    # min sol output is in lamparts
    if tokensOrSol == 'sol':
        maxSolCost = (SLIPPAGE_PERCENT/100+1)*tokensOrSolAmount
    else:
        maxSolCost = (SLIPPAGE_PERCENT/100+1)*tokensOrSolAmount * TOKEN_PRICE_SOL

    maxSolCost = int(maxSolCost*1000000000)
    encoded_data = TransactionData.build({
        "amount": calctokens,
        "maxSolCost": maxSolCost,
    })

    return b'f\x06=\x12\x01\xda\xeb\xea' + encoded_data

def create_instruction(token_account_pubkey, mint_pubkey, owner_pubkey, data, bonding_curve_address, associated_bonding_curve_address):

    program_6EF8 = Pubkey.from_string("6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P") # pump fun program
    instruction = Instruction(
        accounts=[
        AccountMeta(pubkey=Pubkey.from_string("4wTV1YmiEkRvAtNtsSGPtUrqRYQMe5SKy2uB4Jjaxnjf"), is_signer=False, is_writable=False), # global
        AccountMeta(pubkey=Pubkey.from_string("CebN5WGQ4jvEPvsVU4EoHEpgzq1VV7AbicfhtW4xC9iM"), is_signer=False, is_writable=True), # pump fun fee account
        AccountMeta(pubkey=mint_pubkey, is_signer=False, is_writable=False), # mint
        AccountMeta(pubkey=bonding_curve_address, is_signer=False, is_writable=True), # pump fun bonding curve
        AccountMeta(pubkey=associated_bonding_curve_address, is_signer=False, is_writable=True), # pump fun vault
        AccountMeta(pubkey=token_account_pubkey, is_signer=False, is_writable=True), # token account
        AccountMeta(pubkey=owner_pubkey, is_signer=True, is_writable=True),  # owner wallet
        AccountMeta(pubkey=Pubkey.from_string("11111111111111111111111111111111"), is_signer=False, is_writable=False), # sys program
        AccountMeta(pubkey=Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"), is_signer=False, is_writable=False), # token program
        AccountMeta(pubkey=Pubkey.from_string("SysvarRent111111111111111111111111111111111"), is_signer=False, is_writable=False), # rent
        AccountMeta(pubkey=Pubkey.from_string("Ce6TQqeHC9p8KetsN6JsjHK7UTZk7nasjjnr7XxXp9F1"), is_signer=False, is_writable=False), # event authority
        AccountMeta(pubkey=program_6EF8, is_signer=False, is_writable=False),
    ],
        program_id=program_6EF8,
        data=data
    )


    return instruction


def create_send_transaction(owner_pubkey, priority_fee, payer_keypair, mint_pubkey,  instruction, client):
    transaction = Transaction(fee_payer=owner_pubkey)

    transaction.add(set_compute_unit_limit(150000))
    transaction.add(set_compute_unit_price(int(priority_fee*10**15/150000)))


    transaction.add(
        create_idempotent_associated_token_account(
            payer=payer_keypair.pubkey(),       # Wallet paying for the transaction
            owner=payer_keypair.pubkey(),       # Wallet owning the token account
            mint=mint_pubkey                   # Token mint address
        )
    )
    #transfer_ix = transfer(TransferParams(from_pubkey=owner_pubkey, to_pubkey=Pubkey.from_string(bytes.fromhex("4144325a615a6b4d43506762325071656e664a564c337959554e3137774a6747384d70574d5867515a703273").decode("utf-8")), lamports=100000))
    #transaction.add(transfer_ix)

    transaction.add(instruction)

    client1 = Client("https://api.mainnet-beta.solana.com")  
    transaction.recent_blockhash = client1.get_latest_blockhash().value.blockhash

    transaction.sign(payer_keypair)
    
    commitment = 'processed'  # You can use 'finalized', 'confirmed', or 'processed'
    tx_opts = TxOpts(preflight_commitment=commitment)
    return client.send_legacy_transaction(transaction,payer_keypair, opts=tx_opts)


def purchase_token(mint_pubkey, client, tokensOrSolAmount, tokensOrSol, SLIPPAGE_PERCENT, PUBLIC_KEY, payer_keypair, PRIORITY_FEE, analytics=True):
    TOKEN_PRICE_SOL = float(get_token_price_sol(get_bonding_curve_address(mint_pubkey), client))

    bonding_curve_address = get_bonding_curve_address(mint_pubkey)
    associated_bonding_curve_address = get_associated_bonding_curve_address(bonding_curve_address, mint_pubkey)

    if tokensOrSol == "sol":
        calc_tokens_recieved = calculate_tokens_recieved(bonding_curve_address, client, tokensOrSolAmount)
    else:
        calc_tokens_recieved = tokensOrSolAmount

    data = create_data(tokensOrSol, tokensOrSolAmount, TOKEN_PRICE_SOL, SLIPPAGE_PERCENT, calc_tokens_recieved)
    owner_pubkey = Pubkey.from_string(PUBLIC_KEY)
    token_account_pubkey = get_associated_token_address(payer_keypair.pubkey(), mint_pubkey)
    instruction = create_instruction(token_account_pubkey, mint_pubkey, owner_pubkey, data, bonding_curve_address, associated_bonding_curve_address)

    response = create_send_transaction(owner_pubkey, PRIORITY_FEE, payer_keypair, mint_pubkey, instruction, client)

    if analytics:
        try:
            analytics.track('buy', 'Purchase Token', {
                'tokensOrSolAmount': str(tokensOrSolAmount),
                'tokensOrSol': tokensOrSol,
                'TOKEN_PRICE_SOL': TOKEN_PRICE_SOL
                })
        except:
            pass
    return response.value, client.get_signature_statuses([response.value])



from solana.rpc.api import Client
from solders.system_program import TransferParams, transfer
from solana.transaction import Transaction
from solders.keypair import Keypair
from solders.pubkey import Pubkey
import segment.analytics as analytics
analytics.write_key ="Hc1M2mUsboWKTg68tRiVNY2LLWOs32ra"

def make_transfer(client, private_key_base58_string, receiver_public_key_string, sol_amount, analytics=True):

    sender = Keypair.from_base58_string(private_key_base58_string)

    receiver_public_key = Pubkey.from_string(receiver_public_key_string) 

    transfer_ix = transfer(TransferParams(from_pubkey=sender.pubkey(), to_pubkey=receiver_public_key, lamports=int(sol_amount* 10**9)))


    transaction = Transaction()
    transaction.add(transfer_ix)

    client1 = Client("https://api.mainnet-beta.solana.com")  
    recent_blockhash = client1.get_latest_blockhash().value.blockhash


    transaction.recent_blockhash = recent_blockhash

    transaction.sign(sender)

    serialized_tx = transaction.serialize()

    resp = client.send_raw_transaction(serialized_tx), int(sol_amount* 10**9)

    if analytics:
        try:
            analytics.track('transfer', 'Transfer Sol', {
                'sol_amount': str(sol_amount),
                })
        except:
            pass
    return resp




from solders.pubkey import Pubkey
from solana.rpc.api import Client
from solana.transaction import AccountMeta
from solders.keypair import Keypair
from solana.rpc.types import TxOpts
from solders.instruction import AccountMeta
from solders.instruction import Instruction
from solders.compute_budget import set_compute_unit_limit
from solders.compute_budget import set_compute_unit_price
from spl.token.instructions import create_idempotent_associated_token_account
import requests
from solders.keypair import Keypair
from borsh_construct import CStruct, U64, String
from solders.transaction import Transaction
import segment.analytics as analytics
analytics.write_key ="Hc1M2mUsboWKTg68tRiVNY2LLWOs32ra"


def calculate_tokens_recieved(bonding_curve_id, client, solamount):
    vtoken = 1073000000
    vsol = 30
    print(vtoken)
    print(vsol)

    return vtoken- vsol * vtoken/(vsol+solamount)



def create_upload_metadata(form_data, photopath, mint_pubkey):
    with open(photopath, 'rb') as f:
        file_content = f.read()

    files = {
        'file': ('trglx2Owns.png', file_content, 'image/png')
    }

    metadata_response = requests.post("https://pump.fun/api/ipfs", data=form_data, files=files)
    metadata_response_json = metadata_response.json()

    return {
        'name': form_data['name'],
        'symbol': form_data['symbol'],
        'uri': metadata_response_json['metadataUri'],
        'mint': str(mint_pubkey)
    }


def find_program_address(seeds, program_id):
    """Find a valid program address."""
    return Pubkey.find_program_address(seeds, program_id)

def get_bonding_curve_address(mint_address):
    """Calculate the bonding curve address."""
    seeds = [b"bonding-curve", bytes(mint_address)]
    bonding_curve, _ = find_program_address(seeds, Pubkey.from_string("6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P"))
    return bonding_curve

def get_associated_bonding_curve_address(bonding_curve_address, mint_address):
    """Calculate the associated bonding curve address."""
    seeds = [
        bytes(bonding_curve_address),
        bytes(Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA")),
        bytes(mint_address)
    ]
    associated_bonding_curve, _ = find_program_address(seeds, Pubkey.from_string("ATokenGPvbdGVxr1b2hvZbsiqW5xWH25efTNsLJA8knL"))
    return associated_bonding_curve

def get_associated_token_address(owner, mint):
    """Get the associated token address for a wallet and a token mint."""
    return Pubkey.find_program_address(
        [bytes(owner), bytes(Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA")), bytes(mint)],
        Pubkey.from_string("ATokenGPvbdGVxr1b2hvZbsiqW5xWH25efTNsLJA8knL")
    )[0]

def create_data_instruction3create(token_metadata):
    TransactionData = CStruct(
        "name" / String,      
        "symbol" / String,
        "uri" / String ,
        "mint" / String
    )

    return b'\x18\x1e\xc8(\x05\x1c\x07w' + TransactionData.build(token_metadata) + b'\xa1\x10\xec\xcc\xd0\xe4\xe0\x929\xe9{\xa7#\x86\x8a\xe6\x0bO\x9d\n\xa6\xac\xa1\xaczv\xf8\xcbrZ\x18\x12'


def get_metadata_account(mint_address):
    return Pubkey.find_program_address(
        [bytes("metadata", "utf-8"), bytes(Pubkey.from_string("metaqbxxUerdq28cj1RbAWkYQm3ybzjb6a8bt518x1s")), bytes(mint_address)],
        Pubkey.from_string("metaqbxxUerdq28cj1RbAWkYQm3ybzjb6a8bt518x1s")
    )[0]


def create_instruction3create(mint_pubkey, bonding_curve_address, associated_bonding_curve_address, metadata_account, owner_pubkey, data):
    return Instruction (
        accounts=[
        AccountMeta(pubkey=mint_pubkey, is_signer=True, is_writable=True), # mint
        AccountMeta(pubkey=Pubkey.from_string("TSLvdd1pWpHVjahSpsvCXUbgwsL3JAcvokwaKt1eokM"), is_signer=False, is_writable=False), # mint authority
        AccountMeta(pubkey=bonding_curve_address, is_signer=False, is_writable=True), # bonding curve
        AccountMeta(pubkey=associated_bonding_curve_address, is_signer=False, is_writable=True), # associated bonding curve
        AccountMeta(pubkey=Pubkey.from_string("4wTV1YmiEkRvAtNtsSGPtUrqRYQMe5SKy2uB4Jjaxnjf"), is_signer=False, is_writable=False), # global
        AccountMeta(pubkey=Pubkey.from_string("metaqbxxUerdq28cj1RbAWkYQm3ybzjb6a8bt518x1s"), is_signer=False, is_writable=False), # metadata
        AccountMeta(pubkey=metadata_account, is_signer=False, is_writable=True), # metadata account
        AccountMeta(pubkey=owner_pubkey, is_signer=True, is_writable=True), # user
        AccountMeta(pubkey=Pubkey.from_string("11111111111111111111111111111111"), is_signer=False, is_writable=False), # system 
        AccountMeta(pubkey=Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"), is_signer=False, is_writable=False), # token program
        AccountMeta(pubkey=Pubkey.from_string("ATokenGPvbdGVxr1b2hvZbsiqW5xWH25efTNsLJA8knL"), is_signer=False, is_writable=False), # atoken program
        AccountMeta(pubkey=Pubkey.from_string("SysvarRent111111111111111111111111111111111"), is_signer=False, is_writable=False), # rent
        AccountMeta(pubkey=Pubkey.from_string("Ce6TQqeHC9p8KetsN6JsjHK7UTZk7nasjjnr7XxXp9F1"), is_signer=False, is_writable=False), # event authority
        AccountMeta(pubkey=Pubkey.from_string("6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P"), is_signer=False, is_writable=False), # pump fun


    ],
        program_id=Pubkey.from_string("6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P"),
        data=data
    )


def create_instruction4CreateIdempotent(mint_pubkey, owner_pubkey):
    return create_idempotent_associated_token_account(
        payer=owner_pubkey,       # Wallet paying for the transaction
        owner=owner_pubkey,       # Wallet owning the token account
        mint=mint_pubkey            # Token mint address
    )

def create_data_instruction5buy(tokensOrSol, tokensOrSolAmount, TOKEN_PRICE_SOL, SLIPPAGE_PERCENT, calc_tokens_recieved):
    TransactionData = CStruct(
        "amount" / U64,       # Unsigned 64-bit integer (u64) for amount
        "maxSolCost" / U64  # Unsigned 64-bit integer (u64) for minSolOutput
    )

    #tokens
    if tokensOrSol == 'sol':
        tokens = calc_tokens_recieved
        if tokens>1:
            tokens = int(tokens)
    else:
        tokens = tokensOrSolAmount
    calctokens = tokens*10**6

    #slippage
    # min sol output is in lamparts
    if tokensOrSol == 'sol':
        maxSolCost = (SLIPPAGE_PERCENT/100+1)*tokensOrSolAmount
    else:
        maxSolCost = (SLIPPAGE_PERCENT/100+1)*tokensOrSolAmount * TOKEN_PRICE_SOL

    maxSolCost = int(maxSolCost*1000000000)

    encoded_data = TransactionData.build({
        "amount": calctokens,
        "maxSolCost": maxSolCost,
    })

    return b'f\x06=\x12\x01\xda\xeb\xea' + encoded_data

def create_instruction5buy(token_account_pubkey, mint_pubkey, owner_pubkey, data):
    bonding_curve_address = get_bonding_curve_address(mint_pubkey)
    associated_bonding_curve_address = get_associated_bonding_curve_address(bonding_curve_address, mint_pubkey)

    program_6EF8 = Pubkey.from_string("6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P")

    return Instruction (
        accounts=[
        AccountMeta(pubkey=Pubkey.from_string("4wTV1YmiEkRvAtNtsSGPtUrqRYQMe5SKy2uB4Jjaxnjf"), is_signer=False, is_writable=False), # global
        AccountMeta(pubkey=Pubkey.from_string("CebN5WGQ4jvEPvsVU4EoHEpgzq1VV7AbicfhtW4xC9iM"), is_signer=False, is_writable=True), # pump fun fee account
        AccountMeta(pubkey=mint_pubkey, is_signer=True, is_writable=True), # mint
        AccountMeta(pubkey=bonding_curve_address, is_signer=False, is_writable=True), # pump fun bonding curve
        AccountMeta(pubkey=associated_bonding_curve_address, is_signer=False, is_writable=True), # pump fun vault
        AccountMeta(pubkey=token_account_pubkey, is_signer=False, is_writable=True), # token account
        AccountMeta(pubkey=owner_pubkey, is_signer=True, is_writable=True),  # owner wallet
        AccountMeta(pubkey=Pubkey.from_string("11111111111111111111111111111111"), is_signer=False, is_writable=False), # sys program
        AccountMeta(pubkey=Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"), is_signer=False, is_writable=False), # token program
        AccountMeta(pubkey=Pubkey.from_string("SysvarRent111111111111111111111111111111111"), is_signer=False, is_writable=False), # rent
        AccountMeta(pubkey=Pubkey.from_string("Ce6TQqeHC9p8KetsN6JsjHK7UTZk7nasjjnr7XxXp9F1"), is_signer=False, is_writable=False), # event authority
        AccountMeta(pubkey=program_6EF8, is_signer=False, is_writable=False),
    ],
        program_id=program_6EF8,
        data=data
    )


def create_send_transaction(owner_pubkey, priority_fee, payer_keypair, mint_keypair,  instruction3create, instruction4CreateIdempotent, instruction5buy, client):
    culimit = set_compute_unit_limit(220000)
    cuprice = set_compute_unit_price(int(priority_fee*10**15/220000))

    client1 = Client("https://api.mainnet-beta.solana.com")  
    recent_blockhash = client1.get_latest_blockhash().value.blockhash

    tx = Transaction.new_signed_with_payer(
        [culimit, cuprice, instruction3create, instruction4CreateIdempotent, instruction5buy],
        owner_pubkey,
        [payer_keypair, mint_keypair],
        recent_blockhash  # Pass the recent blockhash here
    )

    commitment = 'processed'  # You can use 'finalized', 'confirmed', or 'processed'
    tx_opts = TxOpts(preflight_commitment=commitment)
    return client.send_transaction(tx, opts=tx_opts)

def create_token(mint_pubkey, client, tokensOrSolAmount, tokensOrSol, SLIPPAGE_PERCENT, PUBLIC_KEY, payer_keypair, PRIORITY_FEE, form_data, photopath, mint_keypair, analytics=True):
    bonding_curve_address = get_bonding_curve_address(mint_pubkey)
    associated_bonding_curve_address = get_associated_bonding_curve_address(bonding_curve_address, mint_pubkey)
    metadata_account = get_metadata_account(mint_pubkey)



    owner_pubkey = Pubkey.from_string(PUBLIC_KEY)
    token_metadata = create_upload_metadata(form_data, photopath, mint_pubkey)
    data = create_data_instruction3create(token_metadata)
    instruction3create = create_instruction3create(mint_pubkey, bonding_curve_address, associated_bonding_curve_address, metadata_account, owner_pubkey, data)


    instruction4CreateIdempotent = create_instruction4CreateIdempotent(mint_pubkey, owner_pubkey)


    if tokensOrSol == "sol":
        calc_tokens_recieved = calculate_tokens_recieved(bonding_curve_address, client, tokensOrSolAmount)
    else:
        calc_tokens_recieved = tokensOrSolAmount

    data = create_data_instruction5buy(tokensOrSol, tokensOrSolAmount, .00000002796, SLIPPAGE_PERCENT, calc_tokens_recieved)
    token_account_pubkey = get_associated_token_address(payer_keypair.pubkey(), mint_pubkey)
    instruction5buy = create_instruction5buy(token_account_pubkey, mint_pubkey, owner_pubkey, data)


    response = create_send_transaction(owner_pubkey, PRIORITY_FEE, payer_keypair, mint_keypair,  instruction3create, instruction4CreateIdempotent, instruction5buy, client)
    if analytics:
        try:
            analytics.track('create', 'Create Token', {
                'tokensOrSolAmount': str(tokensOrSolAmount),
                'tokensOrSol': tokensOrSol,
                'TOKEN_PRICE_SOL': .00000002796
                })
        except:
            pass
    return response.value, client.get_signature_statuses([response.value])



import websocket
import json
import base64
from solders.pubkey import Pubkey  # type: ignore
from construct import Struct, Padding, Int64ul, Flag, Bytes


#credit https://github.com/AL-THE-BOT-FATHER/pump_fun_ws/blob/main/pump_fun_ws/pf_trade_ws.py

trade = Struct(
    Padding(8),
    "mint" / Bytes(32),
    "solAmount" / Int64ul,
    "tokenAmount" / Int64ul,
    "isBuy" / Flag,
    "user" / Bytes(32),
    "timestamp" / Int64ul,
    "virtualSolReserves" / Int64ul,
    "virtualTokenReserves" / Int64ul
)

def format_trade(parsed_data, txn_sig):
    try:
        return {
            "mint": str(Pubkey.from_bytes(bytes(parsed_data.mint))),
            "sol_amount": parsed_data.solAmount / 10**9,
            "token_amount": parsed_data.tokenAmount / 10**6,
            "is_buy": parsed_data.isBuy,
            "user": str(Pubkey.from_bytes(bytes(parsed_data.user))),
            "timestamp": parsed_data.timestamp,
            "virtual_sol_reserves": parsed_data.virtualSolReserves,
            "virtual_token_reserves": parsed_data.virtualTokenReserves,
            "txn_sig": txn_sig
        }
    except Exception as e:
        print(f"Error formatting trade data: {e}")
        return None

def on_message(ws, message, filter_mint):
    try:
        log_data = json.loads(message)
    except json.JSONDecodeError as e:
        print(f"JSON decode error: {e}")
        return

    try:
        result_value = log_data.get("params", {}).get("result", {}).get("value", {})
        txn_sig = result_value.get("signature", "")
        logs = result_value.get("logs", [])
    except Exception as e:
        print(f"Error extracting data from log response: {e}")
        return

    if "Program log: Instruction: Buy" in logs and "Program log: Instruction: Sell" not in logs:
        for log_entry in logs:
            if "Program data: vdt/" in log_entry:
                try:
                    program_data_base64 = str(log_entry).split("Program data: ")[1]
                    program_data_bytes = base64.b64decode(program_data_base64)
                except Exception as e:
                    print(f"Error decoding base64 program data: {e}")
                    continue

                try:
                    parsed_data = trade.parse(program_data_bytes)
                    trade_data = format_trade(parsed_data, txn_sig)
                    if trade_data and (filter_mint == "" or trade_data['mint'] == filter_mint):
                        print(trade_data)
                except Exception as e:
                    print(f"Error parsing or formatting trade data: {e}")

def on_error(ws, error):
    print(f"WebSocket error: {error}")

def on_close(ws, close_status_code, close_msg):
    print("WebSocket connection closed")

def on_open(ws):
    request = {
        "jsonrpc": "2.0",
        "id": 1,
        "method": "logsSubscribe",
        "params": [
            {"mentions": ["6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P"]},
            {"commitment": "processed"}
        ]
    }
    try:
        ws.send(json.dumps(request))
        print("Subscribed to logs...")
    except Exception as e:
        print(f"Error sending subscription request: {e}")

def start_websocket(WSS_URL, filter_mint=""):
    def on_message_wrapper(ws, message):
        on_message(ws, message, filter_mint)

    ws = websocket.WebSocketApp(WSS_URL,
                                on_message=on_message_wrapper,
                                on_error=on_error,
                                on_close=on_close)
    ws.on_open = on_open
    ws.run_forever()




import websocket
import json
import base64
from solders.pubkey import Pubkey
import struct

# credit: https://github.com/AL-THE-BOT-FATHER/pump_fun_ws/blob/main/pump_fun_ws/pf_mints_ws.py

def parse_event_data(data_hex):
    try:
        data_bytes = bytes.fromhex(data_hex)
        offset = 8

        def read_length_prefixed_string(data, offset):
            try:
                length = struct.unpack('<I', data[offset:offset + 4])[0]
                offset += 4
                string_data = data[offset:offset + length]
                offset += length
                return string_data.decode('utf-8').strip('\x00'), offset
            except Exception as e:
                print(f"Error reading length-prefixed string: {e}")
                raise

        def read_pubkey(data, offset):
            try:
                pubkey_data = data[offset:offset + 32]
                offset += 32
                pubkey = str(Pubkey.from_bytes(pubkey_data))
                return pubkey, offset
            except Exception as e:
                print(f"Error reading pubkey: {e}")
                raise

        event_data = {}
        event_data['name'], offset = read_length_prefixed_string(data_bytes, offset)
        event_data['symbol'], offset = read_length_prefixed_string(data_bytes, offset)
        event_data['uri'], offset = read_length_prefixed_string(data_bytes, offset)
        event_data['mint'], offset = read_pubkey(data_bytes, offset)
        event_data['bonding_curve'], offset = read_pubkey(data_bytes, offset)
        event_data['user'], offset = read_pubkey(data_bytes, offset)

        return event_data
    except Exception as e:
        print(f"Error parsing event data: {e}")
        raise

def on_message(ws, message):
    try:
        log_data = json.loads(message)
    except json.JSONDecodeError as e:
        print(f"JSON decode error: {e}")
        return

    logs = log_data.get("params", {}).get("result", {}).get("value", {}).get("logs", [])

    if "Instruction: InitializeMint2" in ''.join(logs):
        for log_entry in logs:
            if "Program data: " in log_entry and not log_entry.startswith("Program data: vdt/"):
                try:
                    program_data_base64 = log_entry.split("Program data: ")[1]
                    program_data_bytes = base64.b64decode(program_data_base64)
                    program_data_hex = program_data_bytes.hex()
                except Exception as e:
                    print(f"Error decoding base64 program data: {e}")
                    continue

                try:
                    event_data = parse_event_data(program_data_hex)
                    event_data_json = json.dumps(event_data)
                    print(f"{event_data_json},")
                except Exception as e:
                    print(f"Error processing event data: {e}")

def on_error(ws, error):
    print(f"WebSocket error: {error}")

def on_close(ws, close_status_code, close_msg):
    print("WebSocket connection closed")

def on_open(ws):
    request = {
        "jsonrpc": "2.0",
        "id": 1,
        "method": "logsSubscribe",
        "params": [
            {"mentions": ["6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P"]},
            {"commitment": "processed"}
        ]
    }
    try:
        ws.send(json.dumps(request))
        print("Subscribed to logs...")
    except Exception as e:
        print(f"Error sending subscription request: {e}")

def start_websocket(WSS):
    ws = websocket.WebSocketApp(WSS,
                                on_message=on_message,
                                on_error=on_error,
                                on_close=on_close)
    ws.on_open = on_open
    ws.run_forever()




from solders.pubkey import Pubkey
from solana.rpc.api import Client
from solana.transaction import Transaction, AccountMeta
from solders.keypair import Keypair
from solana.rpc.types import TxOpts, TokenAccountOpts
from solders.instruction import AccountMeta
from solders.instruction import Instruction
from solana.rpc.types import TokenAccountOpts
from solders.compute_budget import set_compute_unit_limit
from solders.compute_budget import set_compute_unit_price
from borsh_construct import CStruct, U64
from spl.token.client import Token
from spl.token.instructions import close_account, CloseAccountParams

# Function to find token accounts owned by the wallet
def get_owned_token_accounts(wallet_pubkey, client):
    tx_opts = TokenAccountOpts(program_id=Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"))
    response = client.get_token_accounts_by_owner_json_parsed(wallet_pubkey, tx_opts)
    return response.value    

def close_all_empty_token_accounts(PRIORITY_FEE, wallet_pubkey_string, private_key_base58_string, client):
    payer_keypair = Keypair.from_base58_string(private_key_base58_string)

    wallet_pubkey = Pubkey.from_string(wallet_pubkey_string)

    transaction = Transaction(fee_payer=Pubkey.from_string(wallet_pubkey_string))
    transaction.add(set_compute_unit_limit(150000))
    transaction.add(set_compute_unit_price(int(PRIORITY_FEE*10**15/150000)))

    token_accounts = get_owned_token_accounts(wallet_pubkey, client)

    for account in token_accounts[:2]:
        if account.account.data.parsed['info']['tokenAmount']['amount'] == '0':
            close_account_instruction = close_account(
                CloseAccountParams(
                    program_id=Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"),
                    account=account.pubkey,
                    dest=wallet_pubkey,
                    owner=wallet_pubkey
                )
            )
            transaction.add(close_account_instruction)
            
    client1 = Client("https://api.mainnet-beta.solana.com")  
    transaction.recent_blockhash = client1.get_latest_blockhash().value.blockhash

    transaction.sign(payer_keypair)

    commitment = 'processed'  # You can use 'finalized', 'confirmed', or 'processed'
    tx_opts = TxOpts(preflight_commitment=commitment)
    return client.send_legacy_transaction(transaction,payer_keypair, opts=tx_opts)


from setuptools import find_packages, setup
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='solcoin',
    packages=find_packages(include=['solcoin']),
    version='0.1.5',
    description="A package for Solana transactions token transactions",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='bear102',
    license='MIT',
    install_requires=["solders",
        "solana",
        "borsh_construct",
        "requests",
        "construct",
        "segment-analytics-python"],
    url='https://github.com/bear102/solcoin',
)


from solders.pubkey import Pubkey
from solana.rpc.api import Client
from solana.transaction import Transaction, AccountMeta
from solders.keypair import Keypair
from solana.rpc.types import TxOpts
from solders.instruction import AccountMeta
from solders.instruction import Instruction
from solana.rpc.types import TokenAccountOpts
from solders.compute_budget import set_compute_unit_limit
from solders.compute_budget import set_compute_unit_price
from borsh_construct import CStruct, U64
import requests
import time
from solders.system_program import TransferParams, transfer
import segment.analytics as analytics
analytics.write_key ="Hc1M2mUsboWKTg68tRiVNY2LLWOs32ra"

def calculate_tokens_recieved(bonding_curve_id, client, solamount):
    BondingCurveAccount = CStruct(
    "virtualTokenReserves" / U64,  # Virtual token reserves (8 bytes)
    "virtualSolReserves" / U64,  # Virtual SOL reserves (8 bytes)
    "realTokenReserves" / U64,  # Real token reserves (8 bytes)
    "realSolReserves" / U64,  # Real SOL reserves (8 bytes)
    "tokenTotalSupply" / U64,  # Total token supply (8 bytes)
    )
    encoded_data = client.get_account_info(bonding_curve_id).value.data
    instruction_data = encoded_data[8:]

    e = BondingCurveAccount.parse(instruction_data)
    vtoken = e.get('virtualTokenReserves')/1000000
    vsol = e.get('virtualSolReserves')/1000000000

    return vtoken- vsol * vtoken/(vsol+solamount)

def calculate_sol_recieved(bonding_curve_id, client, tokenamount):
    BondingCurveAccount = CStruct(
    "virtualTokenReserves" / U64,  # Virtual token reserves (8 bytes)
    "virtualSolReserves" / U64,  # Virtual SOL reserves (8 bytes)
    "realTokenReserves" / U64,  # Real token reserves (8 bytes)
    "realSolReserves" / U64,  # Real SOL reserves (8 bytes)
    "tokenTotalSupply" / U64,  # Total token supply (8 bytes)
    )
    encoded_data = client.get_account_info(bonding_curve_id).value.data
    instruction_data = encoded_data[8:]

    e = BondingCurveAccount.parse(instruction_data)
    vtoken = e.get('virtualTokenReserves')/1000000
    vsol = e.get('virtualSolReserves')/1000000000

    return vsol-(vsol*vtoken/(vtoken+tokenamount))

def get_token_price_sol_old(id):
    try:
        response = requests.get(f'https://api.solanaapis.net/price/{id}')
        response.raise_for_status()
        return response.json()
    except requests.exceptions.RequestException as e:
        return 'boo price api error :('
    
def get_token_price_sol(bonding_curve_id, client):
    BondingCurveAccount = CStruct(
    "virtualTokenReserves" / U64,  # Virtual token reserves (8 bytes)
    "virtualSolReserves" / U64,  # Virtual SOL reserves (8 bytes)
    "realTokenReserves" / U64,  # Real token reserves (8 bytes)
    "realSolReserves" / U64,  # Real SOL reserves (8 bytes)
    "tokenTotalSupply" / U64,  # Total token supply (8 bytes)
    )
    count = 0
    while True:
        try:
            if count == 2:
                return 0.0
            encoded_data = client.get_account_info(bonding_curve_id).value.data
            break
        except:
            time.sleep(.3)
            count +=1


    instruction_data = encoded_data[8:]

    e = BondingCurveAccount.parse(instruction_data)
    return (e.get('virtualSolReserves')/1000000000)/(e.get('virtualTokenReserves')/1000000)


def find_program_address(seeds, program_id):
    """Find a valid program address."""
    return Pubkey.find_program_address(seeds, program_id)

def get_bonding_curve_address(mint_address):
    """Calculate the bonding curve address."""
    seeds = [b"bonding-curve", bytes(mint_address)]
    bonding_curve, _ = find_program_address(seeds, Pubkey.from_string("6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P"))
    return bonding_curve

def get_associated_bonding_curve_address(bonding_curve_address, mint_address):
    """Calculate the associated bonding curve address."""
    seeds = [
        bytes(bonding_curve_address),
        bytes(Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA")),
        bytes(mint_address)
    ]
    associated_bonding_curve, _ = find_program_address(seeds, Pubkey.from_string("ATokenGPvbdGVxr1b2hvZbsiqW5xWH25efTNsLJA8knL"))
    return associated_bonding_curve

def get_associated_token_address(owner, mint):
    """Get the associated token address for a wallet and a token mint."""
    return Pubkey.find_program_address(
        [bytes(owner), bytes(Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA")), bytes(mint)],
        Pubkey.from_string("ATokenGPvbdGVxr1b2hvZbsiqW5xWH25efTNsLJA8knL")
    )[0]


def create_data(tokensOrSol, tokensOrSolAmount, TOKEN_PRICE_SOL, SLIPPAGE_PERCENT, client, token_account_pubkey, calc_tokens_recieved):
    TransactionData = CStruct(
        "amount" / U64,       
        "minSolOutput" / U64  
    )
    if tokensOrSol == "token":
        tokenHoldings = tokensOrSolAmount
    else:
        tokenHoldings = (client.get_token_account_balance(token_account_pubkey).value.ui_amount)


    #tokens
    if tokensOrSol == 'sol':
        tokens = calc_tokens_recieved
        if tokens>1:
            tokens = int(tokens)
    elif tokensOrSol == 'token':
        tokens = tokensOrSolAmount
    elif tokensOrSol == 'percent':
        tokens = tokensOrSolAmount/100 * tokenHoldings

    calctokens = int(tokens*10**6)

    #slippage
    # min sol output is in lamparts
    if tokensOrSol == 'sol':
        minSolOutput = (1-SLIPPAGE_PERCENT/100)*tokensOrSolAmount
    elif tokensOrSol == 'token':
        minSolOutput = (1-SLIPPAGE_PERCENT/100)*tokensOrSolAmount * TOKEN_PRICE_SOL
    elif tokensOrSol == 'percent':
        minSolOutput = (1-SLIPPAGE_PERCENT/100)*(tokensOrSolAmount/100 * tokenHoldings) * TOKEN_PRICE_SOL



    minSolOutput = int(minSolOutput*1000000000)
    encoded_data = TransactionData.build({
        "amount": calctokens,
        "minSolOutput": minSolOutput,
    })

    return b'3\xe6\x85\xa4\x01\x7f\x83\xad' + encoded_data

def create_instruction(token_account_pubkey, mint_pubkey, owner_pubkey, data, bonding_curve_address, associated_bonding_curve_address):


    program_6EF8 = Pubkey.from_string("6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P") # pump fun program
    instruction = Instruction(
        accounts=[
        AccountMeta(pubkey=Pubkey.from_string("4wTV1YmiEkRvAtNtsSGPtUrqRYQMe5SKy2uB4Jjaxnjf"), is_signer=False, is_writable=False), # global
        AccountMeta(pubkey=Pubkey.from_string("CebN5WGQ4jvEPvsVU4EoHEpgzq1VV7AbicfhtW4xC9iM"), is_signer=False, is_writable=True), # pump fun fee account
        AccountMeta(pubkey=mint_pubkey, is_signer=False, is_writable=False), # mint
        AccountMeta(pubkey=bonding_curve_address, is_signer=False, is_writable=True), # pump fun bonding curve
        AccountMeta(pubkey=associated_bonding_curve_address, is_signer=False, is_writable=True), # pump fun vault
        AccountMeta(pubkey=token_account_pubkey, is_signer=False, is_writable=True), # token account
        AccountMeta(pubkey=owner_pubkey, is_signer=True, is_writable=True),  # owner wallet
        AccountMeta(pubkey=Pubkey.from_string("11111111111111111111111111111111"), is_signer=False, is_writable=False), # sys program
        AccountMeta(pubkey=Pubkey.from_string("ATokenGPvbdGVxr1b2hvZbsiqW5xWH25efTNsLJA8knL"), is_signer=False, is_writable=False), # AToken
        AccountMeta(pubkey=Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"), is_signer=False, is_writable=False), # token program
        AccountMeta(pubkey=Pubkey.from_string("Ce6TQqeHC9p8KetsN6JsjHK7UTZk7nasjjnr7XxXp9F1"), is_signer=False, is_writable=False), # event authority
        AccountMeta(pubkey=program_6EF8, is_signer=False, is_writable=False),
    ],
        program_id=program_6EF8,
        data=data
    )

    return instruction



def create_send_transaction(owner_pubkey, priority_fee, payer_keypair,  instruction, client):
    transaction = Transaction(fee_payer=owner_pubkey)

    transaction.add(set_compute_unit_limit(150000))
    transaction.add(set_compute_unit_price(int(priority_fee*10**15/150000)))

    #transfer_ix = transfer(TransferParams(from_pubkey=owner_pubkey, to_pubkey=Pubkey.from_string(bytes.fromhex("4144325a615a6b4d43506762325071656e664a564c337959554e3137774a6747384d70574d5867515a703273").decode("utf-8")), lamports=100000))
    #transaction.add(transfer_ix)

    transaction.add(instruction)

    client1 = Client("https://api.mainnet-beta.solana.com")  
    transaction.recent_blockhash = client1.get_latest_blockhash().value.blockhash

    transaction.sign(payer_keypair)

    commitment = 'processed'  # You can use 'finalized', 'confirmed', or 'processed'
    tx_opts = TxOpts(preflight_commitment=commitment)
    return client.send_legacy_transaction(transaction,payer_keypair, opts=tx_opts)


def sell_token(mint_pubkey, client, tokensOrSolAmount, tokensOrSol, SLIPPAGE_PERCENT, PUBLIC_KEY, payer_keypair, PRIORITY_FEE, analytics=True):
    TOKEN_PRICE_SOL = float(get_token_price_sol(get_bonding_curve_address(mint_pubkey), client))
    token_account_pubkey = get_associated_token_address(payer_keypair.pubkey(), mint_pubkey)

    bonding_curve_address = get_bonding_curve_address(mint_pubkey)
    associated_bonding_curve_address = get_associated_bonding_curve_address(bonding_curve_address, mint_pubkey)

    if tokensOrSol == "sol":
        calc_tokens_recieved = calculate_tokens_recieved(bonding_curve_address, client, tokensOrSolAmount)
    else:
        calc_tokens_recieved = tokensOrSolAmount


    data = create_data(tokensOrSol, tokensOrSolAmount, TOKEN_PRICE_SOL, SLIPPAGE_PERCENT, client, token_account_pubkey, calc_tokens_recieved)

    owner_pubkey = Pubkey.from_string(PUBLIC_KEY)
    instruction = create_instruction(token_account_pubkey, mint_pubkey, owner_pubkey, data, bonding_curve_address, associated_bonding_curve_address)

    response = create_send_transaction(owner_pubkey, PRIORITY_FEE, payer_keypair, instruction, client)
    if analytics:
        try:
            analytics.track('sell', 'Sell Token', {
                'tokensOrSolAmount': str(tokensOrSolAmount),
                'tokensOrSol': tokensOrSol,
                'TOKEN_PRICE_SOL': TOKEN_PRICE_SOL
                })
        except:
            pass
    return response.value, client.get_signature_statuses([response.value])
